/*
 * Portions of this software are copyrighted by the 
 * 
 *     University of South Carolina,
 *     Sun Microsystems, Inc., 
 * 
 * and other parties. 
 * 
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
 * DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
 * IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
 * NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 */


package edu.sc.seis.TauP;

import java.util.Vector;
import java.io.*;

/**
  * provides storage and methods for distance, time and tau increments for
  * a branch. A branch is a group of layers bounded by discontinuities or
  * reversals in slowness gradient.
  *
  * @version 1.0  Tue Feb  2 20:48:34 GMT 1999
  * @author H. Philip Crotwell
  *
  */
public class TauBranch implements Serializable, Cloneable {

    /** Turns on debugging output. */
    transient public boolean DEBUG = false;

    /** The type of wave for this branch, P or S. */
    protected boolean isPWave;
	
    /** The minimum depth of this layer. */
    protected double topDepth;

    /** The maximum depth of this layer. */
    protected double botDepth;
   
    /** The maximum ray parameter that can penetrate into this branch. 
     *  Time, distance, and tau increments are undefined (0.0) for ray
     *  parameters larger than this. */
    protected double maxRayParam;
   
    /** The minimum ray parameter that is turned, but not reflected, in
     *  this  branch. If no rays turn, then it is set
     *  equal to maxRayParam. */
    protected double minTurnRayParam;

    /** The minimum ray parameter that is turned or critically 
     *  reflected in this branch. If no rays turn, then it is set
     *  equal to maxRayParam. */
    protected double minRayParam;

    /** Holds distance (radians) evaluated at the ith ray parameter for 
     *  this branch. */
    protected double[] dist;
   
    /** Holds time (seconds) evaluated at the ith ray parameter for 
     *  this branch. */
    protected double[] time;
   
    /** Holds tau evaluated at the ith ray parameter for this branch. */
    protected double[] tau;

    // Constructors --------------------------------------------------------

    public TauBranch(double topDepth, double botDepth, boolean isPWave) {
	this.topDepth = topDepth;
	this.botDepth = botDepth;
	this.isPWave = isPWave;
    }
	
    // Methods -------------------------------------------------------------

    //Accessor methods

    /** @returns the minimum (top) depth of this layer. */
    public double getTopDepth() {
	return topDepth;
    }

    /** @returns the maximum (bottom) depth of this layer. */
    public double getBotDepth() {
	return botDepth;
    }

    /** @returns the maximum ray parameter that can penetrate into this 
     *  branch. Time, distance, and tau increments are undefined, set 
     *  to 0.0, for ray parameters larger than this. */
    public double getMaxRayParam() {
	return maxRayParam;
    }

    /** @returns the minimum ray parameter that is turned, but not reflected,
     *  in this branch.
     *  */
    public double getMinTurnRayParam() {
	return minTurnRayParam;
    }

    /** @returns the minimum ray parameter that is turned or reflected in this 
     *  branch.
     *  */
    public double getMinRayParam() {
	return minRayParam;
    }

    /** @returns an array, cloned, containing distance (radians) evaluated at
     *  the i_th ray parameter for this branch. */
    public double[] getDist() {
	return (double[])dist.clone();
    }

    /** @returns an array, cloned, containing time (seconds) evaluated at the 
     *  i_th ray parameter for this branch. */
    public double[] getTime() {
	return (double[])time.clone();
    }

    /** @returns an array, cloned, containing tau (seconds) evaluated at the 
     *  i_th ray parameter for this branch. */
    public double[] getTau() {
	return (double[])tau.clone();
    }

    // normal methods
   
    /** Calculates tau for this branch, between slowness layers
     * topLayerNum and botLayerNum, inclusive. 
     *  @exception NoSuchLayerException if a needed slowness layer
     *     cannot be found.
     *  @exception SlownessModelException if there is a problem with the
     *     slowness model
     *  @exception TauModelException if the slownessmodel and taumodel are
     *     not compatible
     */
    public void createBranch(SlownessModel sMod, 
			     double minPSoFar, 
			     double rayParams[])
	throws NoSuchLayerException, SlownessModelException, TauModelException
    {
	int layerNum;
	TimeDist timeDist;
	SlownessLayer layer;
	double p;


	int topLayerNum = sMod.layerNumberBelow(topDepth, isPWave);
	int botLayerNum = sMod.layerNumberAbove(botDepth, isPWave);
	SlownessLayer topSLayer = sMod.getSlownessLayerClone(topLayerNum, isPWave);
	SlownessLayer botSLayer = sMod.getSlownessLayerClone(botLayerNum, isPWave);

	if (topSLayer.topDepth != topDepth || botSLayer.botDepth != botDepth) {
	    if (topSLayer.topDepth != topDepth &&
		Math.abs(topSLayer.topDepth - topDepth) < 0.000001) {
		// really close, so just move top
		System.err.println("Changing topDepth: "+
				   "\ntopDepth: "+topDepth+" "+topSLayer.topDepth+
				   "\nbotDepth: "+botDepth+" "+botSLayer.botDepth);
		topDepth = topSLayer.topDepth;
	    } else if (botSLayer.botDepth != botDepth &&
		       Math.abs(botSLayer.botDepth - botDepth) < 0.000001) {
		// really close, so just move bottom
		System.err.println("Changing botDepth: "+
				   "\ntopDepth: "+topDepth+" "+topSLayer.topDepth+
				   "\nbotDepth: "+botDepth+" "+botSLayer.botDepth);
		botDepth = botSLayer.botDepth;
	    } else {
		// bad match, throw exception
		throw new TauModelException(
					    "createBranch: TauBranch not compatible with slowness sampling:"+
					    "\ntopDepth: "+topDepth+" "+topSLayer.topDepth+
					    "\nbotDepth: "+botDepth+" "+botSLayer.botDepth);
	    }
	}
	
	/* Here we set minTurnRayParam to be the ray parameter that turns
	 * within the layer, not including total reflections off of the
	 * bottom. maxRayParam is the largest ray parameter that can penetrate
	 * this branch. minRayParam is the minimum ray parameter that turns or
	 * is totally reflected in this branch. */
	maxRayParam = minPSoFar;
	minTurnRayParam = sMod.getMinTurnRayParam(botDepth, isPWave);
	minRayParam = sMod.getMinRayParam(botDepth, isPWave);
	
	tau = new double[rayParams.length];
	dist = new double[rayParams.length];
	time = new double[rayParams.length];
      
	for (int rayNum=0;rayNum<rayParams.length;rayNum++) {
	    p = rayParams[rayNum];
	    timeDist = calcTimeDist(sMod, topLayerNum, botLayerNum, p);
	    dist[rayNum] = timeDist.dist;
	    time[rayNum] = timeDist.time;
	    tau[rayNum] = time[rayNum] - p * dist[rayNum];
	    if (DEBUG && (rayNum % ((int)rayParams.length/10) == 0)) {
		System.out.print(rayNum+", ");
	    }
	}
    }

    /** calculates the time and distance increments for the given
     *  ray parameter. The topDepth and botDepth must be correct as
     *  they determine the bounds on the integration/summing. 
     *  @exception NoSuchLayerException if topLayerNum or botLayerNum
     *     are not in the slowness model.
     *  @exception SlownessModelException if the ray with ray parameter p
     *		 turns within a layer instead of at the bottom.
     */
    public TimeDist calcTimeDist(SlownessModel sMod, int topLayerNum,
				 int botLayerNum, double p)
	throws NoSuchLayerException, SlownessModelException
    {
	int layerNum;
	TimeDist timeDist = new TimeDist(p);
	SlownessLayer layer;
 
	if (p <=maxRayParam) {
	    layerNum = topLayerNum;
	    layer = sMod.getSlownessLayer(layerNum, isPWave);

	    while (layerNum<=botLayerNum && p<=layer.topP && p<=layer.botP) {
		timeDist.add(sMod.layerTimeDist(p,layerNum, isPWave));
		layerNum++;
		if (layerNum<=botLayerNum) {
		    layer=sMod.getSlownessLayerClone(layerNum, isPWave);
		}
	    }
	    if ((layer.topP-p)*(p-layer.botP)>0) {
		throw new SlownessModelException(
						 "Ray turns in the middle of this"+
						 " layer. layerNum = "+ layerNum+
						 " sphericalRayParam "+p+
						 " layer ="+layer);
	    }
	}
	return timeDist;
    }
	
    /** Inserts the distance, time, and tau increment for the slowness 
     *  sample given to the branch. This is used for making the depth
     *  correction to a tau model for a non-surface source. 
     *  @throws TauModelException if the tau branch is not compatable with the
     *     slowness sampling
     *  @see edu.sc.seis.TauP.TauModel.depthCorrect(double) */
    protected void insert(double rayParam, SlownessModel sMod, int index)
	throws NoSuchLayerException,
	SlownessModelException,
	TauModelException
    {
	int topLayerNum = sMod.layerNumberBelow(topDepth, isPWave);
	int botLayerNum = sMod.layerNumberAbove(botDepth, isPWave);
	SlownessLayer topSLayer = sMod.getSlownessLayerClone(topLayerNum, isPWave);
	SlownessLayer botSLayer = sMod.getSlownessLayerClone(botLayerNum, isPWave);

	if (topSLayer.topDepth != topDepth || botSLayer.botDepth != botDepth) {
	    throw new TauModelException(
					"insert: TauBranch depths not compatible with slowness sampling:"+
					"\ntopDepth: "+topDepth+" "+topSLayer.topDepth+
					"\nbotDepth: "+botDepth+" "+botSLayer.botDepth);
	}

	TimeDist td = new TimeDist(rayParam, 0.0, 0.0);
	TimeDist temptd;
     	if (topSLayer.botP >= rayParam && topSLayer.topP >= rayParam) {
	    for (int i=topLayerNum;i<=botLayerNum;i++) {
		if (sMod.getSlownessLayer(i,isPWave).botP < rayParam) {
	         		// so we don't sum below the turning depth
		    break;
		} else {
		    temptd = sMod.layerTimeDist(rayParam, i, isPWave);
		    td.dist += temptd.dist;
		    td.time += temptd.time;
		}
	    }
	}
      
	shiftBranch(index);
	dist[index] = td.dist;
	time[index] = td.time;
	tau[index] = td.time - rayParam * td.dist;

    }

    /** generates a new tau branch by "subtracting" the given tau branch from
     * this tau branch. The given tau branch is assumed to by the upper part
     * of this branch.
     * @argument indexP specifies where a new ray coresponding to a P wave sample
     * has been added, it is -1 if no ray parameter has been added to topBranch.
     * @arguement indexS is similar to indexP except for a S wave sample. 
     * Note that although the ray parameters for indexP and indexS were for
     * the P and S waves that turned at the source depth, both ray parameters 
     * need to be added to both P and S branches. */
    protected TauBranch difference(TauBranch topBranch, 
				   int indexP, 
				   int indexS,
				   SlownessModel sMod, 
				   double minPSoFar, 
				   double rayParams[]) 
	throws NoSuchLayerException, 
	SlownessModelException,
	TauModelException 
    {
	if (topBranch.topDepth != topDepth || topBranch.botDepth > botDepth) {
	    if (topBranch.topDepth != topDepth &&
		Math.abs(topBranch.topDepth - topDepth) < 0.000001) {
		// really close, so just move top
		topDepth = topBranch.topDepth;
	    } else {
		// bad match, throw exception
		throw new TauModelException(
					    "createBranch: TauBranch not compatible with slowness sampling:"+
					    "\ntopDepth: "+topDepth+" "+topBranch.topDepth+
					    "\nbotDepth: "+botDepth+" "+topBranch.botDepth);
	    }
	}
	
	if (topBranch.isPWave != isPWave) {
	    throw new TauModelException("Can't difference branches: "+
					"topBranch.topDepth="+topBranch.topDepth +
					" topDepth="+topDepth+" topBranch.botDepth="+topBranch.botDepth+
					" botDepth="+botDepth+" waveTypes:"+topBranch.isPWave+" "+isPWave);
	}

	// find the top and bottom slowness layers of bottom half
	int topLayerNum = sMod.layerNumberBelow(topBranch.botDepth, isPWave);
	int botLayerNum = sMod.layerNumberAbove(botDepth, isPWave);
	SlownessLayer topSLayer = sMod.getSlownessLayer(topLayerNum, isPWave);
	SlownessLayer botSLayer = sMod.getSlownessLayer(botLayerNum, isPWave);

	if (topSLayer.topDepth != topBranch.botDepth || 
	    botSLayer.botDepth != botDepth) {
	    throw new TauModelException(
					"difference: TauBranch not compatible with slowness sampling:"+
					"\ntopDepth: "+topBranch.botDepth+" "+topSLayer.topDepth+
					"\nbotDepth: "+botDepth+" "+botSLayer.botDepth+
					"\n"+topSLayer+"\n"+botSLayer);
	}

	// make sure indexP and indexS really correspond to
	// new ray parameters at the top of this branch
	SlownessLayer sLayer = sMod.getSlownessLayer(
						     sMod.layerNumberBelow(topBranch.botDepth, true), true);
	if (indexP >= 0 && sLayer.topP != rayParams[indexP]) {
	    throw new TauModelException(
					"difference: P wave index doesn't match top layer.\n "+
					sMod.layerNumberBelow(topBranch.botDepth, true)+
					"\n"+rayParams[indexP-1]+
					"\n"+rayParams[indexP]+
					"\n"+rayParams[indexP+1]+
					"\nP="+sLayer+"\nS="+
					sMod.getSlownessLayer(
							      sMod.layerNumberBelow(topBranch.botDepth, 
										    false),
							      false));
	}
	sLayer = sMod.getSlownessLayer(
				       sMod.layerNumberBelow(topBranch.botDepth, false), false);
	if (indexS >= 0 && sLayer.topP != rayParams[indexS]) {
	    throw new TauModelException(
					"difference: S wave index doesn't match top layer. "+
					sMod.layerNumberBelow(topBranch.botDepth, false)+
					" "+rayParams[indexS-1]+" "+rayParams[indexS]+
					" "+rayParams[indexS+1]+
					"\n"+sLayer);
	}
	sLayer = null;
	
	// construct the new TauBranch, going from the bottom of 
	// the top half to the bottom of the whole branch
	TauBranch botBranch = new TauBranch(topBranch.botDepth,
					    botDepth, 
					    isPWave);
	botBranch.maxRayParam = topBranch.minRayParam;
	botBranch.minTurnRayParam = minTurnRayParam;
	botBranch.minRayParam = minRayParam;
	
	double PRayParam=-1.0, SRayParam=-1.0;
	TimeDist timeDistP = new TimeDist();
	TimeDist timeDistS = new TimeDist();
	int arrayLength = dist.length;
	if (indexP != -1) {
	    arrayLength++;
	    PRayParam = rayParams[indexP];
	    timeDistP = botBranch.calcTimeDist(sMod, 
					       topLayerNum, 
					       botLayerNum, 
					       PRayParam);
	}
	if (indexS != -1 && indexS != indexP) {
	    arrayLength++;
	    SRayParam = rayParams[indexS];
	    timeDistS = botBranch.calcTimeDist(sMod, 
					       topLayerNum, 
					       botLayerNum, 
					       SRayParam);
	} else {
	    // in case indexS==indexP then we only need one
	    indexS = -1;
	}
	
	// allocate enough space
	botBranch.dist = new double[arrayLength];
	botBranch.time = new double[arrayLength];
	botBranch.tau = new double[arrayLength];
	
	if (indexP == -1) {
	    // then both are -1 so no new ray parameters added
	    for (int i=0;i<dist.length;i++) {
		botBranch.dist[i] = dist[i] - topBranch.dist[i];
		botBranch.time[i] = time[i] - topBranch.time[i];
		botBranch.tau[i] = tau[i] - topBranch.tau[i];
	    }
	} else {
	    if (indexS == -1) {
		// only indexP != 0
		for (int i=0;i<indexP;i++) {
		    botBranch.dist[i] = dist[i] - topBranch.dist[i];
		    botBranch.time[i] = time[i] - topBranch.time[i];
		    botBranch.tau[i] = tau[i] - topBranch.tau[i];
		}
	      
		botBranch.dist[indexP] = timeDistP.dist;
		botBranch.time[indexP] = timeDistP.time;
		botBranch.tau[indexP] =  timeDistP.time - 
		    PRayParam*timeDistP.dist;
	      
		for (int i=indexP;i<dist.length;i++) {
		    botBranch.dist[i+1] = dist[i] - topBranch.dist[i+1];
		    botBranch.time[i+1] = time[i] - topBranch.time[i+1];
		    botBranch.tau[i+1] = tau[i] - topBranch.tau[i+1];
		}
	    } else {
		// both indexP and indexS != -1 so we have two new samples
		for (int i=0;i<indexS;i++) {
		    botBranch.dist[i] = dist[i] - topBranch.dist[i];
		    botBranch.time[i] = time[i] - topBranch.time[i];
		    botBranch.tau[i] = tau[i] - topBranch.tau[i];
		}

		botBranch.dist[indexS] = timeDistS.dist;
		botBranch.time[indexS] = timeDistS.time;
		botBranch.tau[indexS] =  timeDistS.time - 
		    SRayParam*timeDistS.dist;
	      
		for (int i=indexS;i<indexP;i++) {
		    botBranch.dist[i+1] = dist[i] - topBranch.dist[i+1];
		    botBranch.time[i+1] = time[i] - topBranch.time[i+1];
		    botBranch.tau[i+1] = tau[i] - topBranch.tau[i+1];
		}
	      
		// put into indexP+1 as we have already shifted by 1 
		// due to indexS
		botBranch.dist[indexP+1] = timeDistP.dist;
		botBranch.time[indexP+1] = timeDistP.time;
		botBranch.tau[indexP+1] =  timeDistP.time - 
		    PRayParam*timeDistP.dist;

		for (int i=indexP;i<dist.length;i++) {
		    botBranch.dist[i+2] = dist[i] - topBranch.dist[i+2];
		    botBranch.time[i+2] = time[i] - topBranch.time[i+2];
		    botBranch.tau[i+2] = tau[i] - topBranch.tau[i+2];
		}
	    }
	}
      
	return botBranch;
    }

    public void shiftBranch(int index) {
	double[] newDist = new double[dist.length+1];
	System.arraycopy(dist, 0, newDist, 0, index);
	newDist[index] = 0.0;
	System.arraycopy(dist, index, newDist, index+1, dist.length-index);
	dist = newDist;

	double[] newTime = new double[time.length+1];
	System.arraycopy(time, 0, newTime, 0, index);
	newTime[index] = 0.0;
	System.arraycopy(time, index, newTime, index+1, time.length-index);
	time = newTime;

	double[] newTau = new double[tau.length+1];
	System.arraycopy(tau, 0, newTau, 0, index);
	newTau[index] = 0.0;
	System.arraycopy(tau, index, newTau, index+1, tau.length-index);
	tau = newTau;
    }
   
    public TimeDist[] path(double rayParam, boolean downgoing,
			   SlownessModel sMod) 
	throws SlownessModelException {
	if (rayParam > maxRayParam) {return null;}
	
	int topLayerNum;
	int botLayerNum;
	try {
	    topLayerNum = sMod.layerNumberBelow(topDepth, isPWave);
	    botLayerNum = sMod.layerNumberAbove(botDepth, isPWave);
	} catch (NoSuchLayerException e) {
	    throw new SlownessModelException(
					     "Caught NoSuchLayerException. This likely means the"+
					     "SlownessModel and TauModel are out of sync. "+
					     e.getMessage());
	}
      
	TimeDist[] thePath = new TimeDist[botLayerNum - topLayerNum +1];
	
	int sLayerNum;
	int pathIndex = 0;
	double turnDepth;
	SlownessLayer sLayer, turnSLayer;

	/** check to make sure layers and branches are compatable. */
	sLayer = sMod.getSlownessLayer(topLayerNum, isPWave);
	if (sLayer.topDepth != topDepth) {
	    throw new SlownessModelException(
					     "Branch and Slowness model are not compatible! "+
					     sLayer.topDepth+" != "+topDepth+"=topDepth");
	}
	sLayer = sMod.getSlownessLayer(botLayerNum, isPWave);
	if (sLayer.botDepth != botDepth) {
	    throw new SlownessModelException(
					     "Branch and Slowness model are not compatible! "+
					     sLayer.botDepth+" != "+botDepth+"=botDepth");
	}
      
	if (downgoing) {
	    sLayerNum = topLayerNum;
	    sLayer = sMod.getSlownessLayer(sLayerNum, isPWave);
	    
	    while (sLayer.botP >= rayParam && sLayerNum <= botLayerNum) {
		if (!sLayer.isZeroThickness()) {
		    thePath[pathIndex] = 
			sMod.layerTimeDist(rayParam, sLayerNum, isPWave);
		    thePath[pathIndex].depth = sLayer.botDepth;
		    pathIndex++;
		}
		sLayerNum++;
		if (sLayerNum <= botLayerNum) {
		    sLayer = sMod.getSlownessLayer(sLayerNum, isPWave);
		}
	    }
	  
	    if (sLayerNum <= botLayerNum && !sLayer.isZeroThickness()) {
		turnDepth = sLayer.bullenDepthFor(rayParam,
						  sMod.getRadiusOfEarth());

 
		turnSLayer = new SlownessLayer(sLayer.topP,
					       sLayer.topDepth, 
					       rayParam, 
					       turnDepth);
	      
		thePath[pathIndex] = turnSLayer.bullenRadialSlowness(
								     rayParam, sMod.getRadiusOfEarth());
		thePath[pathIndex].depth = turnSLayer.botDepth;
		pathIndex++;
	    }
	} else {
	    // upgoing
	    sLayerNum = botLayerNum;
	    sLayer = sMod.getSlownessLayer(sLayerNum, isPWave);
	    while (sLayer.topP <= rayParam || sLayer.isZeroThickness()) {
		sLayerNum--;
		sLayer = sMod.getSlownessLayer(sLayerNum, isPWave);
	    }
	    
	    if (sLayer.botP < rayParam) {
		turnDepth = sLayer.bullenDepthFor(rayParam,
						  sMod.getRadiusOfEarth());
		turnSLayer = new SlownessLayer(sLayer.topP,
					       sLayer.topDepth, 
					       rayParam, turnDepth);
		
		thePath[pathIndex] = turnSLayer.bullenRadialSlowness(
								     rayParam, sMod.getRadiusOfEarth());
		thePath[pathIndex].depth = turnSLayer.topDepth;
		pathIndex++;
		sLayerNum--;
		if (sLayerNum >= topLayerNum) {
		    sLayer = sMod.getSlownessLayer(sLayerNum, isPWave);
		}
	    }
	    
	    while (sLayerNum >= topLayerNum) {
		if (!sLayer.isZeroThickness()) {
		    thePath[pathIndex] = 
			sMod.layerTimeDist(rayParam, sLayerNum, isPWave);
		    thePath[pathIndex].depth = sLayer.topDepth;
		    pathIndex++;
		}
		sLayerNum--;
		if (sLayerNum >= topLayerNum) {
		    sLayer = sMod.getSlownessLayer(sLayerNum, isPWave);
		}
	    }
 
	}
	TimeDist[] tempPath = new TimeDist[pathIndex];
	System.arraycopy(thePath,0, tempPath,0 ,pathIndex);
	return tempPath;
    }
    
    public void writeToStream(DataOutputStream dos) throws IOException {
	dos.writeInt(getClass().getName().length());
	dos.writeBytes(getClass().getName());
	
	dos.writeDouble(topDepth);
	dos.writeDouble(botDepth);
	dos.writeDouble(maxRayParam);
	dos.writeDouble(minRayParam);
	dos.writeDouble(minTurnRayParam);
	
	dos.writeInt(dist.length);
	for (int i=0;i<dist.length;i++) {
	    dos.writeDouble(dist[i]);
	}
	dos.writeInt(time.length);
	for (int i=0;i<time.length;i++) {
	    dos.writeDouble(time[i]);
	}
	dos.writeInt(tau.length);
	for (int i=0;i<tau.length;i++) {
	    dos.writeDouble(tau[i]);
	}
    }
    
    public static TauBranch readFromStream(DataInputStream dis) 
	throws IOException, ClassNotFoundException, InstantiationException,
	IllegalAccessException {
	int length;
       
	byte[] classString = new byte[dis.readInt()];
	dis.read(classString);
	Class tBranchClass = Class.forName(new String(classString));
	TauBranch tBranch = (TauBranch)tBranchClass.newInstance();
      
	tBranch.topDepth = dis.readDouble();
	tBranch.botDepth = dis.readDouble();
	tBranch.maxRayParam = dis.readDouble();
	tBranch.minRayParam = dis.readDouble();
	tBranch.minTurnRayParam = dis.readDouble();
      
	length = dis.readInt();
	tBranch.dist = new double[length];
	for (int i=0;i<tBranch.dist.length;i++) {
	    tBranch.dist[i] = dis.readDouble();
	}
	length = dis.readInt();
	tBranch.time = new double[length];
	for (int i=0;i<tBranch.time.length;i++) {
	    tBranch.time[i] = dis.readDouble();
	}
	length = dis.readInt();
	tBranch.tau = new double[length];
	for (int i=0;i<tBranch.tau.length;i++) {
	    tBranch.tau[i] = dis.readDouble();
	}
	return tBranch;
    }
   
    /**
     * Returns a clone of this TauBranch object. Note that super.clone()
     * handles all normal variables while the arrays need to be cloned
     * separately to generate a new array as opposed to a new reference
     * to the old array. 
     * @see Cloneable
     */
    public Object clone() {
	TauBranch newObject;
	try {
	    newObject = (TauBranch)super.clone();
	    newObject.dist = (double[])dist.clone();
	    newObject.time = (double[])time.clone();
	    newObject.tau = (double[])tau.clone();
	    return newObject;
	} catch (CloneNotSupportedException e) {
	    // Can't happen, but...
	    System.err.println("Caught CloneNotSupportedException: "+
			       e.getMessage());
	    throw new InternalError(e.toString());
	}
    }

    public String toString() {
	String desc = "Tau Branch\n";
	desc += " topDepth = "+topDepth+"\n";
	desc += " botDepth = "+botDepth+"\n";
	desc += " maxRayParam="+maxRayParam+" minTurnRayParam="+minTurnRayParam;
	desc += " minRayParam="+minRayParam+"\n";

	/*
	  for (int i=0;i<tau.length;i++) {
	  desc += "i = "+i+" time="+time[i]+" dist="+dist[i]+" tau="+tau[i]+"\n";
	  }
	*/
	return desc;
    }
   
}
