/*
 * Portions of this software are copyrighted by the 
 * 
 *     University of South Carolina,
 *     Sun Microsystems, Inc., 
 * 
 * and other parties. 
 * 
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
 * DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
 * IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
 * NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 */


/**
 * package for storage and manipulation of seismic earth models.
 *
 */
package edu.sc.seis.TauP;


import java.io.*;
import java.util.Vector;

/**
  * The VelocityModelLayer class stores and manipulates a singly layer.
  * An entire velocity model is implemented as an Vector of layers.
  *
  * @version 1.0  Tue Feb  2 20:48:34 GMT 1999
  * @author H. Philip Crotwell
 */
public class VelocityLayer 
   implements Cloneable, Serializable
{
   private int myLayerNumber;
   public double topDepth, botDepth;
   public double topPVelocity, botPVelocity;
   public double topSVelocity, botSVelocity;
   public double topDensity=2.6, botDensity=2.6;
   public double topQp=1000, botQp=1000;
   public double topQs=2000, botQs=2000;

   public VelocityLayer() {
      this.myLayerNumber = 0;
   }

   public VelocityLayer(int myLayerNumber) {
      this.myLayerNumber = myLayerNumber;
   }

   public Object clone() {
      try {
         VelocityLayer newObject = (VelocityLayer)super.clone();
         return newObject;
      } catch (CloneNotSupportedException e) {
	         // Cannot happen, we support clone
	         // and our parent is Object, which supports clone.
         throw new InternalError(e.toString());
      }
   }

   public double evaluateAtBottom(char materialProperty) 
      throws NoSuchMatPropException
   {
      double answer;

      switch (materialProperty) {
         case 'P':
         case 'p':
            answer = botPVelocity;
            break;
         case 's':
         case 'S':
            answer = botSVelocity;
            break;
         case 'r': case 'R':
         case 'D': case 'd':
            answer = botDensity;
            break;
         default:
            throw new NoSuchMatPropException(materialProperty);
      }
      return answer;
   }

   public double evaluateAtTop(char materialProperty) 
      throws NoSuchMatPropException
   {
      double answer;

      switch (materialProperty) {
         case 'P':
         case 'p':
            answer = topPVelocity;
            break;
         case 's':
         case 'S':
            answer = topSVelocity;
            break;
         case 'r': case 'R':
         case 'D': case 'd':
            answer = topDensity;
            break;
         default:
            throw new NoSuchMatPropException(materialProperty);
      }
      return answer;
   }

   public double evaluateAt(double depth, char materialProperty) 
      throws NoSuchMatPropException
   {
      double slope, answer;

      switch (materialProperty) {
         case 'P':
         case 'p':
            slope = (botPVelocity - topPVelocity)/
                     (botDepth - topDepth);
            answer = slope*(depth - topDepth)+topPVelocity;
            break;
         case 's':
         case 'S':
            slope = (botSVelocity - topSVelocity)/
                     (botDepth - topDepth);
            answer = slope*(depth - topDepth)+topSVelocity;
            break;
         case 'r': case 'R':
         case 'D': case 'd':
            slope = (botDensity - topDensity)/
                     (botDepth - topDepth);
            answer = slope*(depth - topDepth)+topDensity;
            break;
         default:
            System.out.println("I don't understand this material property: " +
                                materialProperty +
                               "\nUse one of P p S s R r D d");
            throw new NoSuchMatPropException(materialProperty);
      }
      return answer;
   }


   public void writeToStream(DataOutputStream dos) throws IOException {
      dos.writeInt(getClass().getName().length());
      dos.writeBytes(getClass().getName());
      dos.writeInt(myLayerNumber);
      dos.writeDouble(topDepth);
      dos.writeDouble(botDepth);
      dos.writeDouble(topPVelocity);
      dos.writeDouble(botPVelocity);
      dos.writeDouble(topSVelocity);
      dos.writeDouble(botSVelocity);
      dos.writeDouble(topDensity);
      dos.writeDouble(botDensity);
      dos.writeDouble(topQp);
      dos.writeDouble(botQp);
      dos.writeDouble(topQs);
      dos.writeDouble(botQs);

   }
   
   public static VelocityLayer readFromStream(DataInputStream dis) 
   throws IOException, ClassNotFoundException, IllegalAccessException,
   InstantiationException {
      int length;
      
      byte[] classString = new byte[dis.readInt()];
      dis.read(classString);
      Class vLayerClass = Class.forName(new String(classString));
      VelocityLayer vLayer = (VelocityLayer)vLayerClass.newInstance();
      
      vLayer.myLayerNumber = dis.readInt();
      vLayer.topDepth = dis.readDouble();
      vLayer.botDepth = dis.readDouble();
      vLayer.topPVelocity = dis.readDouble();
      vLayer.botPVelocity = dis.readDouble();
      vLayer.topSVelocity = dis.readDouble();
      vLayer.botSVelocity = dis.readDouble();
      vLayer.topDensity = dis.readDouble();
      vLayer.botDensity = dis.readDouble();
      vLayer.topQp = dis.readDouble();
      vLayer.botQp = dis.readDouble();
      vLayer.topQs = dis.readDouble();
      vLayer.botQs = dis.readDouble();
      
      return vLayer;
   }
   
   public String toString() {
      String description;

      description = myLayerNumber + " " + topDepth + " " + botDepth;
      description += " P " + topPVelocity + " " + botPVelocity;
      description += " S " + topSVelocity + " " + botSVelocity;
      description += " Density " + topDensity + " " + botDensity;
      return description;
   }
}
